from enum import Enum


class InteractionType(Enum):
    """
    Enum representing the types of interactions that can be recorded.
    """
    INITIAL = "initial"
    METHOD_CALL = "method_call"
    ATTRIBUTE_ACCESS = "attribute_access"
    ATTRIBUTE_ASSIGNMENT = "attribute_assignment"


import json
import os
from typing import Any, Dict, List, Set, Tuple, Union
from unittest.mock import Mock

GUMMIES_TYPE = "__gummies#type__"


def serialize_exception(e: Exception) -> str:
    """
    Serialize an exception into a JSON string.

    Args:
        e: The exception to be serialized.

    Returns:
        The serialized exception as a JSON string.
    """
    return json.dumps({"type": type(e).__name__, "message": str(e)})


def deserialize_exception(serialized_exception: str) -> Exception:
    """
    Deserialize a JSON string into an exception.

    Args:
        serialized_exception: The serialized exception as a JSON string.

    Returns:
        The deserialized exception.
    """
    exception_dict = json.loads(serialized_exception)
    # Dynamically create an exception of the correct type with the same message
    return eval(exception_dict["type"])(exception_dict["message"])


def to_serializable_interactions(obj):
    """
    Convert a object back to a serializable object form

    Args:
        obj: The object to be converted.

    Returns:
        The converted object.
    """
    if isinstance(obj, (int, float, str, bool, type(None))):
        return obj
    if isinstance(obj, list):
        return [to_serializable_interactions(item) for item in obj]
    if isinstance(obj, set):
        return {to_serializable_interactions(item) for item in obj}
    if isinstance(obj, tuple):
        return tuple(to_serializable_interactions(item) for item in obj)
    if isinstance(obj, dict):
        return {key: to_serializable_interactions(value) for key, value in obj.items()}
    if isinstance(obj, bytes):
        return {GUMMIES_TYPE: "bytes", "data": str(obj, encoding="utf-8")}
    if isinstance(obj, Exception):
        return {
            GUMMIES_TYPE: "exception",
            "classname": obj.__class__.__name__,
            "module": obj.__class__.__module__,
            "data": serialize_exception(obj),
        }
    return {
        GUMMIES_TYPE: "mock",
        "classname": obj.__class__.__name__,
        "module": obj.__class__.__module__,
    }


def from_serializable_interactions(obj: Any) -> Any:
    """
    Convert a serializable object back to its original form.

    Args:
        obj: The object to be converted.

    Returns:
        The converted object.
    """
    if isinstance(obj, list):
        return [from_serializable_interactions(item) for item in obj]
    if isinstance(obj, set):
        return {from_serializable_interactions(item) for item in obj}
    if isinstance(obj, tuple):
        return tuple(from_serializable_interactions(item) for item in obj)
    if isinstance(obj, dict):
        if obj.get(GUMMIES_TYPE) == "bytes":
            return bytes(obj.get("data"), encoding="utf-8")
        if obj.get(GUMMIES_TYPE) == "exception":
            return deserialize_exception(obj.get("data"))
        if obj.get(GUMMIES_TYPE) == "mock":
            return Mock()
        return {
            key: from_serializable_interactions(value) for key, value in obj.items()
        }
    return obj


def get_json_or_file_path(s: str) -> Union[Dict[str, Any], str, None]:
    """
    Check if the input string is a valid JSON or a valid file path.

    Args:
        s: The input string to be checked.

    Returns:
        If the input string is a valid JSON, return the JSON object and the type 'dict'.
        If the input string is a valid file path, return the file path and the type 'str'.
        If the input string is neither a valid JSON nor a valid file path, return None.
    """
    # Check if s is a valid JSON
    try:
        return json.loads(s), dict
    except json.JSONDecodeError:
        pass

    # Check if s is a valid file path
    if os.path.isfile(s):
        return s, str

    return None, None



import json
import os
from typing import Any, Optional

from rehearser.constants import InteractionType
from rehearser.utils import to_serializable_interactions

INTERACTIONS_RAW_FILE_DEFAULT_DIRECTORY = "./raw_files/rehearser_proxy/"


class RehearserProxy:
    """
    A proxy class that records all interactions with the proxied object.
    """

    def __init__(self, obj: Any) -> None:
        """
        Initialize a RehearserProxy object.

        Args:
            obj: The object to be proxied.
        """
        self._interactions_file_directory = None
        self._interactions_file_name = None
        self._interactions_file_path = None
        self._obj = obj
        self._interactions = []
        ## Record initial state of a obj
        for attr_name in obj.__dict__.keys():
            if attr_name not in [
                "_obj",
                "_interactions",
                "_interactions_file_directory",
                "_interactions_file_name",
                "_interactions_file_path",
            ]:
                self._interactions.append(
                    {
                        "type": InteractionType.INITIAL.name,
                        "name": attr_name,
                        "result": getattr(obj, attr_name),
                    }
                )

    def set_interactions_file_directory(self, directory: str) -> None:
        """
        Set the directory of the interactions file.

        Args:
            directory: The directory of the interactions file.
        """
        self._interactions_file_directory = directory

    def get_interactions_file_directory(self) -> str:
        """
        Get the directory of the interactions file.

        Returns:
            The directory of the interactions file.
        """
        return self._interactions_file_directory

    def __getattr__(self, name: str) -> Any:
        """
        Override the __getattr__ method to record interactions.

        Args:
            name: The name of the attribute being accessed.

        Returns:
            A wrapper function that records the interaction and then calls the original method.
        """
        # print(f"[temp_debug][RehearserProxy]Triggered!! {name}")
        attr = getattr(self._obj, name)
        if callable(attr):

            def wrapper(*args, **kwargs):
                interaction = {
                    "type": InteractionType.METHOD_CALL.name,
                    "name": name,
                    "args": args,
                    "kwargs": kwargs,
                }
                try:
                    result = attr(*args, **kwargs)
                    interaction["result"] = result
                    return result
                except Exception as e:
                    interaction["exception"] = e
                    raise
                finally:
                    self._interactions.append(interaction)

            return wrapper
        else:
            value = getattr(self._obj, name)
            self._interactions.append(
                {
                    "type": InteractionType.ATTRIBUTE_ACCESS.name,
                    "name": name,
                    "result": value,
                }
            )
            return value

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Override the __setattr__ method to record interactions.

        Args:
            name: The name of the attribute being set.
            value: The value being assigned to the attribute.
        """
        if name in [
            "_obj",
            "_interactions",
            "_interactions_file_directory",
            "_interactions_file_name",
            "_interactions_file_path",
        ]:
            super().__setattr__(name, value)
        else:
            self._interactions.append(
                {
                    "type": InteractionType.ATTRIBUTE_ASSIGNMENT.name,
                    "name": name,
                    "value": value,
                }
            )
            setattr(self._obj, name, value)

    def __delattr__(self, name: str) -> None:
        """
        Override the __delattr__ method to record interactions.

        Args:
            name: The name of the attribute being deleted.
        """
        self._interactions.append(
            {"type": InteractionType.ATTRIBUTE_DELETION.name, "name": name}
        )
        delattr(self._obj, name)

    def get_interactions(self) -> list[dict]:
        """
        Get the interactions recorded so far.

        Returns:
            A list of interactions.
        """
        return self._interactions

    def get_interactions_serializable_json(self) -> list[dict]:
        """
        Get the interactions recorded so far in serializable json format.

        Returns:
            A list of interactions in serializable json format.
        """
        return to_serializable_interactions(self._interactions)

    def get_interactions_file_path(self) -> str:
        """
        Get the path of the interactions file.

        Returns:
            The path of the interactions file.
        """
        if not self._interactions_file_path:
            directory = self._interactions_file_directory or INTERACTIONS_RAW_FILE_DEFAULT_DIRECTORY
            class_name = self._obj.__class__.__name__
            directory += f"{class_name}/"
            filename = self._interactions_file_name or "latest_interactions.json"
            self._interactions_file_path = os.path.join(directory, filename)
        return self._interactions_file_path

    def write_interactions_to_file(self, file_path: Optional[str] = None) -> None:
        """
        Write the interactions recorded so far to a file.

        Args:
            file_path: The path of the file to write the interactions to.
        """
        if file_path:
            self._interactions_file_path = file_path
        else:
            file_path = self.get_interactions_file_path()
        doc_str = json.dumps(self.get_interactions_serializable_json(), indent=2)
        print(f"Writing interactions to file: {file_path}, with {len(doc_str)} chars")
        directory = os.path.dirname(file_path)
        os.makedirs(directory, exist_ok=True)
        with open(file_path, "w") as f:
            f.write(doc_str)



from setuptools import setup, find_packages

setup(
    name="rehearser",
    version="0.1.5",
    url="https://github.com/kevinchwong/rehearser",
    author="Kevin C. Wong",
    author_email="kevinchwong@gmail.com",
    description="Rehearser makes writing reliable unit tests super easy!",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    keywords='unit testing, contract testing, testing tools, open source, developer tools, rehearsal, proxy, rehearser',

)


import json
from typing import Any, Callable, Dict, List, Union
from unittest.mock import MagicMock, Mock, PropertyMock

from rehearser.constants import InteractionType
from rehearser.utils import (from_serializable_interactions,
                             get_json_or_file_path)


class MockGenerator:
    def __init__(
        self, interactions_src: Union[str, Dict[str, Any], List[Dict[str, Any]]]
    ):
        """
        Initialize the MockGenerator with interactions source.

        Args:
            interactions_src: The source of interactions, either a JSON string or a dictionary.
        """
        self._method_interactions = {}
        self._attribute_interactions = {}
        if isinstance(interactions_src, str):
            interactions_src, src_type = get_json_or_file_path(interactions_src)
            if not src_type:
                return
            if src_type == str:
                with open(interactions_src) as f:
                    interactions_src = json.load(f)

        interactions = from_serializable_interactions(interactions_src)
        for interaction in interactions:
            if interaction["type"] == InteractionType.METHOD_CALL.name:
                if interaction["name"] not in self._method_interactions:
                    self._method_interactions[interaction["name"]] = []
                self._method_interactions[interaction["name"]].append(interaction)
            elif interaction["type"] == InteractionType.ATTRIBUTE_ACCESS.name:
                if interaction["name"] not in self._attribute_interactions:
                    self._attribute_interactions[interaction["name"]] = []
                self._attribute_interactions[interaction["name"]].append(interaction)

    def create_mock(self):
        """
        Create a mock object based on the interactions.

        Returns:
            A mock object with methods and attributes set based on the interactions.
        """
        mock = MagicMock()

        # all methods share the same interactions log
        attr_side_effect = {}
        for attr_name, interactions in self._attribute_interactions.items():
            attr_side_effect[attr_name] = []
            for attr_info in interactions:
                attr_side_effect[attr_name].append(attr_info.get("result"))
        for attr_name in self._attribute_interactions.keys():
            print(
                f"attr_name: {attr_name}, attr_side_effect: {attr_side_effect[attr_name]}"
            )
            setattr(
                type(mock),
                attr_name,
                PropertyMock(side_effect=attr_side_effect[attr_name]),
            )

        # all methods share the same interactions log
        for method_name, interactions in self._method_interactions.items():
            setattr(
                mock, method_name, Mock(side_effect=self._get_side_effect(interactions))
            )
        return mock

    def _get_side_effect(self, interactions: List[Dict[str, Any]]) -> Callable:
        """
        Get the side effect function for a mock method.

        Args:
            interactions: The list of interactions for the method.

        Returns:
            A function that can be used as the side effect for the mock method.
        """

        def side_effect(*args: Any, **kwargs: Any) -> Any:
            """
            The side effect function for the mock method.

            Args:
                *args: The positional arguments for the method call.
                **kwargs: The keyword arguments for the method call.

            Returns:
                The result of the method call, if it was recorded in the interactions.
            """
            if not interactions:
                raise StopIteration
            interaction = interactions.pop(0)
            if "result" in interaction:
                return interaction["result"]
            elif "exception" in interaction:
                raise interaction["exception"]

        return side_effect
